import { __assign, __read, __spreadArray, __values } from "tslib";
import * as React from 'react';
import { observable, autorun, untracked } from 'mobx';
import { remove, set as lodashSet, get as lodashGet, isEqual } from 'lodash-es';
import { EventProxy } from './eventProxy';
import { runtime } from './adapter';
import { generateDataContext } from './common';
import { Repeater } from './constants';
export var widgetKeys = [
    'getConfig',
    'findWidgets',
    'getWidgetsByType',
    'extends',
    'addWidgetsListener',
    'removeWidgetsListener',
    'getOwnerWidget',
    'getDom',
    'domRef',
    '_getInstanceRef',
    '_scope',
];
export var WIDGET_API_MAP = widgetKeys.reduce(function (map, key) {
    map[key] = true;
    return map;
}, {});
export var WidgetsContext = React.createContext({ parent: null });
export function checkVisible(_a) {
    var _visible = _a._visible;
    return _visible !== false && _visible !== '';
}
export function getDom(element, options) {
    if (!element) {
        console.warn('getDom 接口未传入有效的 element');
        return Promise.resolve({});
    }
    var result = {};
    if (options.id)
        result.id = element.id;
    if (options.dataset)
        result.dataset = element.dataset;
    if (options.rect) {
        var rect = element.getBoundingClientRect();
        Object.assign(result, {
            left: rect.left,
            right: rect.right,
            top: rect.top,
            bottom: rect.bottom,
        });
    }
    if (options.size) {
        var rect = element.getBoundingClientRect();
        Object.assign(result, {
            width: rect.width,
            height: rect.height,
        });
    }
    if (options.scrollOffset) {
        Object.assign(result, {
            scrollLeft: element.scrollLeft,
            scrollTop: element.scrollTop,
        });
    }
    if (options.properties && Array.isArray(options.properties)) {
        options.properties.forEach(function (propName) {
            result[propName] = element.getAttribute(propName);
        });
    }
    if (options.computedStyle && Array.isArray(options.computedStyle)) {
        var computedStyle_1 = runtime.getComputedStyle(element);
        options.computedStyle.forEach(function (propName) {
            result[propName] = computedStyle_1[propName];
        });
    }
    return Promise.resolve(result);
}
export function findWidgets(widget, parentType, filterFn, includeInvisibleDescendants) {
    if (!widget)
        return [];
    var _a = widget.children, children = _a === void 0 ? [] : _a;
    var matched = children.filter(filterFn);
    // 过滤掉不可见后代
    if (!includeInvisibleDescendants) {
        children = children.filter(function (item) { return checkVisible(item); });
    }
    children.forEach(function (child) {
        // 如果递归过程中发现了自己，则停止递归
        matched = matched.concat(findWidgets(child, parentType, filterFn, includeInvisibleDescendants));
    });
    // 过滤掉不可见的匹配项
    if (!includeInvisibleDescendants) {
        matched = matched.filter(function (item) { return checkVisible(item); });
    }
    return matched;
}
var retryQueue = [];
// 递归执行
export function retryDataBinds(tryTime) {
    if (tryTime === void 0) { tryTime = 10; }
    var fn = retryQueue.shift();
    if (!fn || tryTime <= 0)
        return;
    try {
        fn({ silent: tryTime > 1 });
    }
    catch (e) {
        console.error('retryDataBinds', e);
    }
    retryDataBinds(tryTime - 1);
}
export function createWidgets(ctx, widgetProps, dataBinds, context) {
    var _a = createWidgetTree(ctx.nodeMap, ctx.widgetProps, widgetProps, dataBinds), nodeTree = _a.nodeTree, nodeMap = _a.nodeMap;
    var widgets = runFor(ctx.widgets, nodeTree, {}, null, null, context);
    return { widgets: widgets, nodeMap: nodeMap };
    /**
     * @param {*} pre widgets
     * @param {*} curForNode
     * @param {*} forItems
     * @param {*} parentLevelWidgets
     * @param {*} parentWidget
     * @returns top level widgets or for dispose
     */
    function runFor(pre, curForNode, forItems, parentLevelWidgets, parentWidget, context) {
        if (!curForNode.value) {
            // 根节点。只有组件才有 value 值
            return createSubTree(curForNode, {});
        }
        /**
         * 组件节点
         */
        var nodeId = curForNode.id;
        var _waForKey = curForNode.value._waForKey;
        var timer = {};
        var dispose = autorun(function (r) {
            var forList = [];
            try {
                clearTimeout(timer[nodeId]);
                /**
                 * untrace 收集context 阶段不添加observer
                 * 消费数据阶段再添加
                 */
                var dataContext = untracked(function () { return generateDataContext(parentWidget); });
                var list = dataBinds[nodeId]._waFor(forItems, undefined, dataContext);
                if (!Array.isArray(list)) {
                    timer[nodeId] = setTimeout(function () {
                        console.warn(nodeId, 'For 循环绑定的数据并不是数组，请检查');
                    }, 1000);
                }
                else {
                    forList = list;
                }
            }
            catch (e) {
                timer[nodeId] = setTimeout(function () {
                    console.error("waFor error ".concat(nodeId), e);
                }, 1000);
            }
            // r.trace();
            // 让 forList 进行监听
            forList.forEach(function () { });
            untracked(function () {
                // dispose widgets before reused instead
                // disposeWidgets(parentLevelWidgets[curForNode.id]);
                var exsitMap = forList.reduce(function (map, item) {
                    if (item === null || item === void 0 ? void 0 : item[_waForKey]) {
                        map[item[_waForKey]] = true;
                    }
                    return map;
                }, {});
                var forWidgets = (parentLevelWidgets === null || parentLevelWidgets === void 0 ? void 0 : parentLevelWidgets[nodeId]) || [];
                var existedWidgetIndex = [];
                var extraWidgetsIndex = [];
                var extraWidgets = [];
                forWidgets.forEach(function (widget, index) {
                    if (exsitMap[widget._key]) {
                        existedWidgetIndex.push(index);
                        // need to use uqique key
                        exsitMap[widget._key] = undefined;
                    }
                    else {
                        extraWidgetsIndex.push(index);
                    }
                });
                var extraWidgetsIndexMap = extraWidgetsIndex.reduce(function (map, item) {
                    map[item] = true;
                    extraWidgets.push(forWidgets[item]);
                    return map;
                }, {});
                // clean nodes of previouse for run
                dfsTree(curForNode, function (node) {
                    var arr = parentLevelWidgets[node.id];
                    remove(arr, function (_, index) {
                        return extraWidgetsIndexMap[index];
                    });
                }, undefined);
                // 清理根 for 的 autorun, 并递归清理子节点
                extraWidgets.map(function (w) {
                    disposeWidgets([w]);
                    parentWidget && remove(parentWidget.children, function (_a) {
                        var id = _a.id;
                        return id === nodeId;
                    });
                });
                forList.forEach(function (item, index) {
                    var _a;
                    var subForItems = __assign(__assign({}, forItems), (_a = {}, _a[nodeId] = item, _a.forItems = __spreadArray([], __read(((forItems === null || forItems === void 0 ? void 0 : forItems['forItems']) || [])), false).concat(item), _a.forIndexes = __spreadArray([], __read(((forItems === null || forItems === void 0 ? void 0 : forItems['forIndexes']) || [])), false).concat(index), _a));
                    createSubTree(curForNode, subForItems /* , context*/);
                });
                // 非初始化时遇到需要重新构建 dataBinds
                retryDataBinds();
            });
        });
        return dispose;
        function createSubTree(curForNode, subForItems /* , context*/) {
            var widgets = {};
            // traverse down the tree to set all widgets
            dfsTree(curForNode, function (node, parentNode, cache) {
                var e_1, _a;
                var _b, _c, _d, _e, _f;
                var parentForWidgetArr = (parentLevelWidgets === null || parentLevelWidgets === void 0 ? void 0 : parentLevelWidgets[node.id]) || [];
                var _waForKey = node.value._waForKey;
                var key = (_b = subForItems[node.id]) === null || _b === void 0 ? void 0 : _b[_waForKey];
                var index = cache[parentNode === null || parentNode === void 0 ? void 0 : parentNode.id]
                    ? cache[parentNode.id].index
                    : parentForWidgetArr.findIndex(function (widget) { return key && widget._key === key; });
                var forExsitWidget = index !== -1 ? parentForWidgetArr[index] : null;
                var preExistWidget = node.exist
                    ? Array.isArray(pre === null || pre === void 0 ? void 0 : pre[node.id])
                        ? lodashGet(pre === null || pre === void 0 ? void 0 : pre[node.id], subForItems.forIndexes)
                        : pre === null || pre === void 0 ? void 0 : pre[node.id]
                    : null;
                var existedWidget = forExsitWidget || preExistWidget || null;
                if (forExsitWidget) {
                    cache[node.id] = {
                        index: index,
                    };
                }
                if (node.forCount === curForNode.forCount) {
                    // 同一层循环作用域内
                    // Leaf node
                    var w_1 = existedWidget;
                    if (!existedWidget) {
                        w_1 = observable(__assign(__assign({}, node.value), { children: [], id: node.id, _key: key, parent: null }));
                        if (node === curForNode) {
                            w_1._disposers = [];
                        }
                        // 更新 scope
                        w_1._scope = {
                            id: node.id,
                            dataContext: context[node.id] || undefined,
                        };
                        if (curForNode.id === node.id) {
                            // 只在 for 绑定的结点上挂 forContext
                            w_1._scope.forContext = (_c = subForItems === null || subForItems === void 0 ? void 0 : subForItems.forItems) === null || _c === void 0 ? void 0 : _c[node.forCount - 1];
                        }
                        w_1._listeners = new EventProxy();
                        // 提供一个给 Node 挂载 API 的方式
                        untracked(function () {
                            if (!w_1.getConfig) {
                                // 兼容非复合组件调用 getConfig 能力
                                w_1.getConfig = function () { return ({}); };
                            }
                            w_1.findWidgets = function (type, includeInvisibleDescendants) {
                                return findWidgets(w_1, w_1.widgetType, type, includeInvisibleDescendants);
                            };
                            w_1.getWidgetsByType = function (type, includeInvisibleDescendants) {
                                return w_1.findWidgets(function (currentWidget) { return currentWidget.widgetType === type; }, includeInvisibleDescendants);
                            };
                            w_1.extends = function (name, fnOrData) {
                                return Object.defineProperty(w_1, name, {
                                    value: fnOrData,
                                    writable: true,
                                });
                            };
                            w_1.addWidgetsListener = function (name, listener) { return w_1._listeners.on(name, listener); };
                            w_1.removeWidgetsListener = function (name, listener) { return w_1._listeners.cancel(name, listener); };
                            w_1.getOwnerWidget = function () { return null; }; // 默认带上一个 widget
                            w_1._getInstanceRef = function () { return null; }; // 默认初始值
                            Object.defineProperty(w_1, '_methods', {
                                get: function () {
                                    var _a;
                                    var instance = w_1._getInstanceRef();
                                    return (_a = instance === null || instance === void 0 ? void 0 : instance.current) === null || _a === void 0 ? void 0 : _a.methods;
                                },
                            });
                        });
                        var parent_1 = parentNode ? widgets[parentNode.id] : parentWidget;
                        if (parent_1) {
                            w_1.parent = parent_1;
                            // 只有可显示 visible 的才存入 children 里
                            if (checkVisible(w_1)) {
                                parent_1.children.push(w_1);
                            }
                        }
                    }
                    else {
                        if (!forExsitWidget && preExistWidget) {
                            if (node.changed) {
                                try {
                                    for (var _g = __values(node.changed), _h = _g.next(); !_h.done; _h = _g.next()) {
                                        var key_1 = _h.value;
                                        w_1[key_1] = node.value[key_1];
                                    }
                                }
                                catch (e_1_1) { e_1 = { error: e_1_1 }; }
                                finally {
                                    try {
                                        if (_h && !_h.done && (_a = _g.return)) _a.call(_g);
                                    }
                                    finally { if (e_1) throw e_1.error; }
                                }
                            }
                            // 仅仅为了受控模式下更新 dataContext
                            if (context === null || context === void 0 ? void 0 : context[w_1.id]) {
                                w_1._scope.dataContext = context[w_1.id];
                            }
                        }
                        disposeWidgets([existedWidget], true);
                    }
                    if (!forExsitWidget) {
                        (_d = parentForWidgetArr.push) === null || _d === void 0 ? void 0 : _d.call(parentForWidgetArr, w_1);
                    }
                    widgets[node.id] = w_1;
                    /**
                     * UserWidget
                     */
                    w_1.__userWidget = new Proxy(new UserWidget(w_1), {
                        get: function (obj, prop) {
                            var _a;
                            if (prop === 'sys') {
                                // 通过命名空间访问内置的方法和属性
                                return obj[prop];
                            }
                            if (prop === 'custom') {
                                // 通过命名空间访问自定义的方法和属性
                                return obj[prop];
                            }
                            // 优先尝试用户自定义的成员（方法和属性）
                            var userCustomMember = (_a = w_1._methods) === null || _a === void 0 ? void 0 : _a[prop];
                            if (userCustomMember)
                                return userCustomMember;
                            // 再尝试内置的方法和属性
                            var buildinMember = obj.sys[prop];
                            if (buildinMember)
                                return buildinMember;
                            // 兼容原来的 widget。最后直接代理到 widget 上，主要是访问组件的属性
                            return w_1[prop];
                        }
                    });
                    /**
                    * Repeater 组件逻辑
                    */
                    if (w_1.widgetType === "".concat(Repeater.MODULE_NAME, ":").concat(Repeater.REPEATER_NAME) || w_1.widgetType === "".concat(Repeater.DEBUG_MODULE_NAME, ":").concat(Repeater.REPEATER_NAME)) {
                        // 该 widget 为 Repeater 组件
                        // 记录该层的 item 别名信息
                        subForItems.repeaterItemAlias = subForItems.repeaterItemAlias || [];
                        subForItems.repeaterItemAlias[node.forCount] = subForItems.repeaterItemAlias[node.forCount] || {};
                        // 这里要考虑同一层多个循环展示组件，不同当前项别名的情况
                        subForItems.repeaterItemAlias[node.forCount]['forIndex_' + w_1.suffix] = w_1.forIndex;
                        subForItems.repeaterItemAlias[node.forCount]['forItem_' + w_1.suffix] = w_1.forItem;
                        // Repeater.items 代理
                        if (!w_1.items) {
                            Object.defineProperty(w_1, 'items', {
                                get: function () {
                                    return (w_1.children || []).map(function (item) {
                                        var descendants = {};
                                        Object.keys((item === null || item === void 0 ? void 0 : item.__descendants) || {}).forEach(function (key) {
                                            descendants[key] = item.__descendants[key].__userWidget;
                                        });
                                        return descendants;
                                    });
                                }
                            });
                        }
                    }
                    if (((_e = parentNode === null || parentNode === void 0 ? void 0 : parentNode.value) === null || _e === void 0 ? void 0 : _e.widgetType) === "".concat(Repeater.MODULE_NAME, ":").concat(Repeater.REPEATER_ITEM_NAME) || ((_f = parentNode === null || parentNode === void 0 ? void 0 : parentNode.value) === null || _f === void 0 ? void 0 : _f.widgetType) === "".concat(Repeater.DEBUG_MODULE_NAME, ":").concat(Repeater.REPEATER_ITEM_NAME)) {
                        node.__inRepeaterScope = true; // Repeater 作用域的标识
                        node.__ancestorId = parentNode.id; // 从 Repeater 的循环虚拟项下的结点开始，结点的循环祖先是循环虚拟项
                    }
                    if (parentNode === null || parentNode === void 0 ? void 0 : parentNode.__inRepeaterScope) {
                        // Repeater 作用域内的所有子孙（排除里面的 Repeater 组件，它本身又产生深一层的作用域）继承父级的循环信息
                        if (!node.__inRepeaterScope) {
                            node.__inRepeaterScope = true;
                            node.__ancestorId = parentNode.__ancestorId;
                        }
                    }
                    if (widgets === null || widgets === void 0 ? void 0 : widgets[node.__ancestorId]) {
                        // 在虚拟项 RepeaterItem 下挂载所有的子孙 __descendants
                        // 这里不要用 lodashSet，否则会出现刷新几次，出现 Error: [mobx.array] Index out of bounds 的错误，原因暂不明
                        widgets[node.__ancestorId].__descendants = widgets[node.__ancestorId].__descendants || {};
                        widgets[node.__ancestorId].__descendants[node.id] = widgets[node.id];
                    }
                    // Setup data binds
                    Object.keys(dataBinds[node.id] || {})
                        .sort(function (a, b) {
                        return a.length - b.length > 0 ? 1 : -1;
                    })
                        .map(function (prop) {
                        if (prop === '_waFor') {
                            return;
                        }
                        var timer = null;
                        function getBindData(options) {
                            if (options === void 0) { options = { silent: false }; }
                            var disposeError = false;
                            var dispose = autorun(function (reaction) {
                                try {
                                    clearTimeout(timer);
                                    /**
                                     * untrace 收集context 阶段不添加observer
                                     * 消费数据阶段再添加
                                     */
                                    var dataContext = untracked(function () { return generateDataContext(w_1); });
                                    // Computed data bind in the next tick since data bind may read widgets data
                                    var value_1 = dataBinds[node.id][prop](subForItems, undefined, dataContext, w_1);
                                    var paths_1 = prop.split('.').filter(function (key) { return !!key; });
                                    if (paths_1.length > 1) {
                                        // 一定要 untracked 不然爆栈了
                                        untracked(function () {
                                            var pre = lodashGet(w_1, prop);
                                            lodashSet(w_1, prop, value_1);
                                            /**
                                             * 深层的对象内部值变化不会触发 obverver，因此浅拷贝对象
                                             */
                                            if (Array.isArray(w_1[paths_1[0]])) {
                                                if (!isEqual(pre, value_1)) {
                                                    w_1[paths_1[0]] = __spreadArray([], __read(w_1[paths_1[0]]), false);
                                                }
                                            }
                                        });
                                    }
                                    else {
                                        // 普通 key 直接赋值
                                        w_1[prop] = value_1;
                                    }
                                    disposeError = false;
                                }
                                catch (e) {
                                    if (prop === '_visible') {
                                        w_1[prop] = false;
                                    }
                                    else {
                                        retryQueue.push(getBindData);
                                        disposeError = true;
                                    }
                                    if (!(options === null || options === void 0 ? void 0 : options.silent)) {
                                        timer = setTimeout(function () {
                                            console.warn("Error computing data bind ".concat(w_1.id, ".").concat(prop), e);
                                            timer = null;
                                        }, 1000);
                                    }
                                }
                            });
                            if (!disposeError && curForNode.id) {
                                widgets[curForNode.id]._disposers.push(dispose);
                            }
                        }
                        getBindData();
                    });
                }
                else {
                    if (existedWidget && forExsitWidget) {
                        // for 项的复用
                        widgets[node.id] = existedWidget;
                    }
                    else if (parentLevelWidgets) {
                        var len = parentLevelWidgets[node.id].push([]);
                        widgets[node.id] = parentLevelWidgets[node.id][len - 1];
                    }
                    else {
                        widgets[node.id] = observable([]);
                    }
                }
            }, undefined);
            // run for of next level
            dfsTree(curForNode, function (node, parentNode) {
                if (node.forCount === curForNode.forCount + 1 && dataBinds[node.id] && dataBinds[node.id]._waFor) {
                    widgets[node.id]._disposers = { dataBinds: [] };
                    var dispose_1 = runFor(pre, node, subForItems, widgets, node.parent && widgets[node.parent.id], context);
                    curForNode.id && widgets[curForNode.id]._disposers.push(dispose_1);
                }
            }, undefined);
            return widgets;
        }
    }
}
function createWidgetTree(preNodes, preWidgets, widgets, dataBinds) {
    var virtualRoot = { children: [], forCount: 0 };
    var nodes = Object.keys(widgets).reduce(function (result, id) {
        var currentValue = widgets[id];
        result[id] = {
            id: id,
            value: currentValue,
            children: [],
            parent: null,
            forCount: 0,
            exist: false,
            changed: [],
        };
        return result;
    }, {});
    // Create widgets tree API
    Object.keys(nodes).map(function (id) {
        var curNode = nodes[id];
        var parent = nodes[widgets[id]._parentId];
        if (!parent) {
            virtualRoot.children.push(curNode);
            return;
        }
        curNode.parent = parent;
        parent.children.push(curNode);
    });
    virtualRoot.children.map(addForCount);
    // dfs, add forCount
    function addForCount(node) {
        var _a;
        if (node.parent) {
            node.forCount = node.parent.forCount;
        }
        if ((_a = dataBinds[node.id]) === null || _a === void 0 ? void 0 : _a._waFor) {
            node.forCount += 1;
        }
        node.children.map(addForCount);
    }
    dfsTree(virtualRoot, function (node, parent) {
        var currentValue = node.value;
        var preValue = preWidgets === null || preWidgets === void 0 ? void 0 : preWidgets[node.id];
        var preNode = preNodes === null || preNodes === void 0 ? void 0 : preNodes[node.id];
        var exist = false;
        var changed = [];
        // 多级for绑定的情况会发生父for不变化，子for更新的情况，此时子for生成会有异常，因此复用只复用一层保证渲染正常
        // https://lcap.cloud.tencent.com/preview/?u=https%3A%2F%2Fci-source-1303824488.cos.ap-shanghai.myqcloud.com%2Flcp%2Fapp-h057yqaw%2Fappdata-1666253755745.json&appId=app-h057yqaw#/nav
        if ((!parent || (parent === null || parent === void 0 ? void 0 : parent.exist)) && preValue && preNode) {
            if (preValue.widgetType === (currentValue === null || currentValue === void 0 ? void 0 : currentValue.widgetType) &&
                preValue._parentId === currentValue._parentId &&
                preNode.forCount === node.forCount) {
                exist = true;
                for (var key in currentValue) {
                    if (!WIDGET_API_MAP[key] && !isEqual(preValue[key], currentValue[key])) {
                        changed.push(key);
                    }
                }
            }
            node.exist = exist;
            node.changed = changed;
        }
    }, undefined);
    return { nodeTree: virtualRoot, nodeMap: nodes };
}
function dfsTree(node, fn, parent, cache, scope) {
    if (cache === void 0) { cache = {}; }
    if (scope === void 0) { scope = { dataContext: {} }; }
    if (node.value) {
        fn(node, parent, cache, scope);
    }
    node.children.map(function (e) { return dfsTree(e, fn, node.value ? node : null, cache, scope); });
}
// dispose autorun
function disposeWidgets(widgets, noRecursive) {
    if (widgets === void 0) { widgets = []; }
    if (noRecursive === void 0) { noRecursive = false; }
    widgets.forEach(function (widget) {
        var disposers = widget._disposers;
        if (disposers) {
            disposers.map(function (dispose) { return dispose(); });
            disposers.splice(0, disposers.length);
        }
        if (!noRecursive) {
            disposeWidgets(widget.children);
        }
    });
}
/**
 * 对外（用户）的 Widget
 */
var UserWidget = /** @class */ (function () {
    function UserWidget(widget) {
        this._widget = null;
        this._widget = widget;
    }
    Object.defineProperty(UserWidget.prototype, "description", {
        get: function () {
            var id = this._widget.id;
            return "\n    \u4F7F\u7528\u8BF4\u660E\uFF1A\n    1. w \u547D\u540D\u7A7A\u95F4\u4E0B\u4E3A\u5185\u7F6E\u65B9\u6CD5\uFF0C\u53EF\u901A\u8FC7 $w.".concat(id, ".w.<\u6210\u5458> \u6216 $w.").concat(id, ".<\u6210\u5458> \u8BBF\u95EE\u3002\n       \u8BBF\u95EE id \u793A\u4F8B: $w.").concat(id, ".id \u6216 $w.").concat(id, ".w.id\n    2. custom \u547D\u540D\u7A7A\u95F4\u4E0B\u4E3A\u7528\u6237\u81EA\u5B9A\u4E49\u5BF9\u5916\u66B4\u9732\u7684\u6210\u5458\uFF0C\u53EF\u901A\u8FC7 $w.").concat(id, ".custom.<\u6210\u5458> \u6216 $w.").concat(id, ".<\u6210\u5458> \u8BBF\u95EE\u3002\n       \u8BBF\u95EE\u81EA\u5B9A\u4E49\u65B9\u6CD5 hello \u793A\u4F8B: $w.").concat(id, ".hello() \u6216 $w.").concat(id, ".custom.hello()\n    3. \u901A\u8FC7 $w.").concat(id, ".<\u6210\u5458> \u8BBF\u95EE\u65F6\uFF0C\u5982\u679C\u81EA\u5B9A\u4E49\u6210\u5458\u4E0E\u5185\u7F6E\u6210\u5458\u91CD\u540D\uFF0C\u5219\u81EA\u5B9A\u4E49\u6210\u5458\u8986\u76D6\u5185\u7F6E\u6210\u5458\u3002\n    4. [\u6CE8\u610F]: \u8BF7\u4E0D\u8981\u76F4\u63A5\u8BBF\u95EE _widget\uFF0C\u5B83\u91CC\u9762\u7684\u6210\u5458\u4E3A\u5185\u90E8\u5B9E\u73B0\uFF0C\u968F\u65F6\u53EF\u80FD\u8FDB\u884C\u8C03\u6574\n    ");
        },
        enumerable: false,
        configurable: true
    });
    Object.defineProperty(UserWidget.prototype, "sys", {
        // 内置属性和方法命名空间
        get: function () {
            var widget = this._widget;
            return {
                /**
                 * 内置属性
                 */
                get id() {
                    return widget.id;
                },
                get name() {
                    return widget.widgetType;
                },
                get parent() {
                    var _a;
                    return (_a = widget.parent) === null || _a === void 0 ? void 0 : _a.__userWidget;
                },
                get children() {
                    var _a;
                    return ((_a = widget.children) === null || _a === void 0 ? void 0 : _a.map(function (item) { return item.__userWidget; })) || [];
                },
                /**
                 * 内置方法
                 */
                closest: function (filter) {
                    var parent = this.parent;
                    if (!filter)
                        return parent;
                    while (parent) {
                        var matched = filter(parent);
                        if (matched)
                            return parent;
                        parent = parent.parent;
                    }
                    return null;
                }
            };
        },
        enumerable: false,
        configurable: true
    });
    Object.defineProperty(UserWidget.prototype, "custom", {
        get: function () {
            return this._widget._methods;
        },
        enumerable: false,
        configurable: true
    });
    return UserWidget;
}());
export { UserWidget };
//# sourceMappingURL=widgets.js.map